/*******************************************************************************
 * Copyright (c) 2017 Nerian Vision Technologies
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *******************************************************************************/

#include "stream/datastream.h"
#include "stream/buffer.h"
#include "misc/infoquery.h"
#include "system/system.h"
#include "device/logicaldevice.h"
#include "device/physicaldevice.h"
#include <algorithm>

namespace GenTL {

DataStream::DataStream(LogicalDevice * logicalDevice, StreamType streamType)
    :Handle(TYPE_STREAM), logicalDevice(logicalDevice), streamType(streamType), framesToAquire(0),
    numDelivered(0), numUnderrun(0), numCaptured(0),
    newBufferEvent(EVENT_NEW_BUFFER, sizeof(S_EVENT_NEW_BUFFER), this),
    opened(false), portImpl(this),
    port("default", "datastream.xml", "StreamPort", "NerianGentTLDataStream", &portImpl) {

}

DataStream::~DataStream() {
    close();
}

Buffer* DataStream::requestBuffer() {
    if(framesToAquire > 0) {
        if(inputPool.size() > 0) {
            return inputPool.front().get();
        } else {
            numUnderrun++;
            return nullptr;
        }
    } else {
        return nullptr;
    }
}

void DataStream::queueOutputBuffer() {
    outputQueue.push_back(inputPool.front());
    inputPool.pop_front();

    if(framesToAquire != GENTL_INFINITE) {
        framesToAquire--;
    }
    numCaptured++;

    // Event notification
    S_EVENT_NEW_BUFFER eventData;
    eventData.BufferHandle = outputQueue.back().get();
    eventData.pUserPointer = outputQueue.back()->getPrivateData();

    newBufferEvent.emitEvent(eventData);
}

template <class T>
bool DataStream::findBuffer(T queue, Buffer* buffer) {
    for(const std::shared_ptr<Buffer> buf: queue) {
        if(buf.get() == buffer) {
            return true;
        }
    }
    return false;
}

GC_ERROR DataStream::announceBuffer(void* pBuffer, size_t iSize, void* pPrivate, BUFFER_HANDLE* phBuffer) {
    if(pBuffer == nullptr || phBuffer == nullptr) {
        return GC_ERR_INVALID_PARAMETER;
    } else if(reinterpret_cast<size_t>(pBuffer) % 16 != 0) {
        // Improperly aligned
        return GC_ERR_INVALID_BUFFER;
    }

    std::unique_lock<std::mutex> lock(logicalDevice->getPhysicalDevice()->lock());

    std::shared_ptr<Buffer> buffer(new Buffer(this, pPrivate,
        reinterpret_cast<unsigned char*>(pBuffer), iSize));
    buffers.push_back(buffer);
    *phBuffer = buffer.get();

    return GC_ERR_SUCCESS;
}

GC_ERROR DataStream::allocAndAnnounceBuffer(size_t iBufferSize, void* pPrivate, BUFFER_HANDLE* phBuffer) {
    if(phBuffer == nullptr) {
        return GC_ERR_INVALID_PARAMETER;
    }

    std::unique_lock<std::mutex> lock(logicalDevice->getPhysicalDevice()->lock());

    std::shared_ptr<Buffer> buffer(new Buffer(this, pPrivate, iBufferSize));
    buffers.push_back(buffer);
    *phBuffer = buffer.get();

    return GC_ERR_SUCCESS;
}

GC_ERROR DataStream::open() {
    if(opened) {
        return GC_ERR_RESOURCE_IN_USE;
    } else {
        opened = true;
        framesToAquire = 0;
        numUnderrun = 0;
        return GC_ERR_SUCCESS;
    }
}

GC_ERROR DataStream::close() {
    if(!opened) {
        return GC_ERR_INVALID_HANDLE;
    } else {
        std::unique_lock<std::mutex> lock(logicalDevice->getPhysicalDevice()->lock());
        inputPool.clear();
        outputQueue.clear();
        opened = false;
        return GC_ERR_SUCCESS;
    }
}

GC_ERROR DataStream::revokeBuffer(BUFFER_HANDLE hBuffer, void ** ppBuffer, void ** ppPrivate) {
    std::unique_lock<std::mutex> lock(logicalDevice->getPhysicalDevice()->lock());
    Buffer* buffer = reinterpret_cast<Buffer*>(hBuffer);

    // Make sure the buffer is not in the input pool or output queue
    if(findBuffer(inputPool, buffer) || findBuffer(outputQueue, buffer)) {
        return GC_ERR_BUSY;
    }

    // Erase buffer from list
    for(auto iter = buffers.begin(); iter != buffers.end(); iter++) {
        if(iter->get() == buffer) {
            if(ppBuffer != nullptr) {
                if(buffer->isConsumerBuffer()) {
                    *ppBuffer = buffer->getData();
                } else {
                    *ppBuffer = nullptr;
                }
            }

            if(ppPrivate != nullptr) {
                *ppPrivate = buffer->getPrivateData();
            }

            buffers.erase(iter);
            return GC_ERR_SUCCESS;
        }
    }

    // Buffer wasn't in the list
    return GC_ERR_INVALID_HANDLE;
}

GC_ERROR DataStream::queueBuffer(BUFFER_HANDLE hBuffer) {
    Buffer* buffer = reinterpret_cast<Buffer*>(hBuffer);
    std::unique_lock<std::mutex> lock(logicalDevice->getPhysicalDevice()->lock());

    // Locate buffer and queueue
    for(const std::shared_ptr<Buffer>& buf: buffers) {
        if(buf.get() == buffer) {
            inputPool.push_back(buf);
            return GC_ERR_SUCCESS;
        }
    }

    return GC_ERR_INVALID_HANDLE;
}

GC_ERROR DataStream::getParentDev(DEV_HANDLE* phDevice) {
    if(phDevice == nullptr) {
        return GC_ERR_INVALID_PARAMETER;
    }

    *phDevice = reinterpret_cast<DEV_HANDLE>(logicalDevice);
    return GC_ERR_SUCCESS;
}

GC_ERROR DataStream::startAcquisition(ACQ_START_FLAGS iStartFlags, uint64_t iNumToAcquire) {
    framesToAquire = iNumToAcquire;
    numDelivered = 0;
    return GC_ERR_SUCCESS;
}

GC_ERROR DataStream::stopAcquisition(ACQ_STOP_FLAGS iStopFlags) {
    framesToAquire = 0;
    return GC_ERR_SUCCESS;
}

GC_ERROR DataStream::flushQueue(ACQ_QUEUE_TYPE iOperation) {
    std::unique_lock<std::mutex> deviceLock(logicalDevice->getPhysicalDevice()->lock());
    bool clearEvents = false;

    switch(iOperation) {
        case ACQ_QUEUE_INPUT_TO_OUTPUT:
            for(unsigned int i=0; i<inputPool.size(); i++) {
                outputQueue.push_back(inputPool[i]);

                S_EVENT_NEW_BUFFER eventData;
                eventData.BufferHandle = inputPool[i].get();
                eventData.pUserPointer = inputPool[i]->getPrivateData();
                newBufferEvent.emitEvent(eventData);
            }
            inputPool.clear();
            break;
        case ACQ_QUEUE_OUTPUT_DISCARD:
            clearEvents = true;
            outputQueue.clear();
            break;
        case ACQ_QUEUE_ALL_TO_INPUT:
            clearEvents = true;
            outputQueue.clear();
            inputPool.clear();
            inputPool.insert(inputPool.end(), buffers.begin(), buffers.end());
            break;
        case ACQ_QUEUE_UNQUEUED_TO_INPUT:
            inputPool.clear();
            for(const std::shared_ptr<Buffer> buf: buffers) {
                if(!findBuffer(outputQueue, buf.get())) {
                    inputPool.push_back(buf);
                }
            }
            break;
        case ACQ_QUEUE_ALL_DISCARD:
            clearEvents = true;
            outputQueue.clear();
            inputPool.clear();
            break;
        default:
            return GC_ERR_NOT_IMPLEMENTED;
    }

    if(clearEvents) {
        newBufferEvent.clearEventQueue();
    }

    return GC_ERR_SUCCESS;
}

GC_ERROR DataStream::getBufferChunkData(BUFFER_HANDLE hBuffer, SINGLE_CHUNK_DATA* pChunkData,
        size_t* piNumChunks) {
    if(piNumChunks == nullptr) {
        return GC_ERR_INVALID_PARAMETER;
    }

    *piNumChunks = 0;
    return GC_ERR_SUCCESS;
}

GC_ERROR DataStream::getBufferID(uint32_t iIndex, BUFFER_HANDLE* phBuffer) {
    if(phBuffer == nullptr) {
        return GC_ERR_INVALID_PARAMETER;
    }

    std::unique_lock<std::mutex> lock(logicalDevice->getPhysicalDevice()->lock());

    if(iIndex >= buffers.size()) {
        return GC_ERR_INVALID_INDEX;
    }

    *phBuffer = buffers[iIndex].get();
    return GC_ERR_SUCCESS;
}

size_t DataStream::getPayloadSizeForStreamType(StreamType streamType) {
    const ImagePair& metaData = logicalDevice->getPhysicalDevice()->getLatestMetaData();
    if(metaData.getWidth() == 0) {
        return 0;
    }

    int pixels = metaData.getWidth() * metaData.getHeight();
    int bytesPixel0 = metaData.getPixelFormat(0) == ImagePair::FORMAT_8_BIT ? 1 : 2;
    int bytesPixel1 = metaData.getPixelFormat(1) == ImagePair::FORMAT_8_BIT ? 1 : 2;
    int imageSize = 1 * pixels;
    int disparitySize = 2 * pixels;
    int image0Size = pixels * bytesPixel0;
    int image1Size = pixels * bytesPixel1;
    int pointCloudSize = pixels*3 * sizeof(float);

    if(streamType == IMAGE_LEFT_STREAM) {
        return imageSize;
    } else if(streamType == IMAGE_RIGHT_STREAM) {
        return imageSize;
    } else if(streamType == DISPARITY_STREAM) {
        return disparitySize;
    } else if(streamType == POINTCLOUD_STREAM) {
        return pointCloudSize;
    } else {
        // Multipart
        return image0Size + image1Size + pointCloudSize;
    }
}

size_t DataStream::getPayloadSize() {
    return getPayloadSizeForStreamType(streamType);
}

uint64_t DataStream::getPixelFormatForStreamType(const ImagePair& metaData, StreamType streamType) {
    if(streamType == IMAGE_LEFT_STREAM || streamType == IMAGE_RIGHT_STREAM) {
        return 0x01080001; //"Mono8"
    } else if(streamType == DISPARITY_STREAM) {
        return 0x01100005; //"Mono12"
    } else if(streamType == POINTCLOUD_STREAM) {
        return 0x026000C0; // "Coord3D_ABC32f"
    } else {
        // Multipart. Querying the pixel format is invalid. We output
        // the mono8 format as a dummy.
        return 0x01080001; //"Mono8"
    }
}

uint64_t DataStream::getPixelFormat(const ImagePair& metaData) {
    return getPixelFormatForStreamType(metaData, streamType);
}

GC_ERROR DataStream::getBufferInfo(BUFFER_HANDLE hBuffer, BUFFER_INFO_CMD iInfoCmd, INFO_DATATYPE* piType,
        void* pBuffer, size_t* piSize) {
    Buffer* buffer = reinterpret_cast<Buffer*>(hBuffer);
    std::unique_lock<std::mutex> lock(logicalDevice->getPhysicalDevice()->lock());
    InfoQuery info(piType, pBuffer, piSize);
    switch(iInfoCmd) {
    case BUFFER_INFO_BASE:
        info.setPtr(buffer->getData());
        break;
    case BUFFER_INFO_SIZE:
        info.setSizeT(buffer->getSize());
        break;
    case BUFFER_INFO_USER_PTR:
        info.setPtr(buffer->getPrivateData());
        break;
    case BUFFER_INFO_TIMESTAMP: {
        int secs = 0, microsecs = 0;
        buffer->getMetaData().getTimestamp(secs, microsecs);
        info.setUInt64(static_cast<uint64_t>(secs)*1000000 + microsecs);
        break;
    }
    case BUFFER_INFO_NEW_DATA:
        info.setBool(findBuffer(outputQueue, buffer));
        break;
    case BUFFER_INFO_IS_QUEUED:
        info.setBool(findBuffer(outputQueue, buffer) || findBuffer(inputPool, buffer));
        break;
    case BUFFER_INFO_IS_ACQUIRING:
        info.setBool(false);
        break;
    case BUFFER_INFO_IS_INCOMPLETE:
        info.setBool(buffer->isIncomplete());
        break;
    case BUFFER_INFO_TLTYPE:
        info.setString("Ethernet");
        break;
    case BUFFER_INFO_SIZE_FILLED:
        if(findBuffer(inputPool, buffer) || buffer->isIncomplete()) {
            info.setSizeT(0);
        } else {
            info.setSizeT(std::min(getPayloadSize(), buffer->getSize()));
        }
        break;
    case BUFFER_INFO_WIDTH:
        info.setSizeT(buffer->getMetaData().getWidth());
        break;
    case BUFFER_INFO_HEIGHT:
        info.setSizeT(buffer->getMetaData().getHeight());
        break;
    case BUFFER_INFO_XOFFSET:
        info.setSizeT(0);
        break;
    case BUFFER_INFO_YOFFSET:
        info.setSizeT(0);
        break;
    case BUFFER_INFO_XPADDING:
        info.setSizeT(0);
        break;
    case BUFFER_INFO_YPADDING:
        info.setSizeT(0);
        break;
    case BUFFER_INFO_FRAMEID:
        info.setUInt64(buffer->getMetaData().getSequenceNumber());
        break;
    case BUFFER_INFO_IMAGEPRESENT:
        info.setBool(true);
        break;
    case BUFFER_INFO_IMAGEOFFSET:
        info.setSizeT(0);
        break;
    case BUFFER_INFO_PAYLOADTYPE:
        info.setSizeT(streamType != MULTIPART_STREAM ? PAYLOAD_TYPE_IMAGE : PAYLOAD_TYPE_MULTI_PART);
        break;
    case BUFFER_INFO_PIXELFORMAT:
        if(streamType == MULTIPART_STREAM) {
            return GC_ERR_NOT_AVAILABLE;
        } else {
            info.setUInt64(getPixelFormat(buffer->getMetaData()));
        }
        break;
    case BUFFER_INFO_PIXELFORMAT_NAMESPACE:
        if(streamType == MULTIPART_STREAM) {
            return GC_ERR_NOT_AVAILABLE;
        } else {
            info.setUInt64(PIXELFORMAT_NAMESPACE_PFNC_32BIT);
        }
        break;
    case BUFFER_INFO_DELIVERED_IMAGEHEIGHT:
        info.setSizeT(buffer->getMetaData().getHeight());
        break;
    case BUFFER_INFO_PIXEL_ENDIANNESS:
        info.setInt(PIXELENDIANNESS_LITTLE);
        break;
    case BUFFER_INFO_DATA_SIZE:
        if(findBuffer(inputPool, buffer)) {
            info.setSizeT(0);
        } else {
            info.setSizeT(getPayloadSize());
        }
        break;
    case BUFFER_INFO_TIMESTAMP_NS: {
        int secs = 0, microsecs = 0;
        buffer->getMetaData().getTimestamp(secs, microsecs);
        info.setUInt64(static_cast<uint64_t>(secs)*1000000000 + static_cast<uint64_t>(microsecs) * 1000);
        break;
    }
    case BUFFER_INFO_DATA_LARGER_THAN_BUFFER:
        info.setBool(getPayloadSize() > buffer->getSize() || buffer->isIncomplete());
        break;
    case BUFFER_INFO_CONTAINS_CHUNKDATA:
        info.setBool(false);
        break;
    default:
        return GC_ERR_NOT_IMPLEMENTED;
    }

    return info.query();

}

GC_ERROR DataStream::getInfo(STREAM_INFO_CMD iInfoCmd, INFO_DATATYPE* piType,
        void* pBuffer, size_t* piSize) {
    InfoQuery info(piType, pBuffer, piSize);
    switch(iInfoCmd) {
        case STREAM_INFO_ID:
            info.setString("default");
            break;
        case STREAM_INFO_NUM_DELIVERED:
            info.setUInt64(numDelivered);
            break;
        case STREAM_INFO_NUM_UNDERRUN:
            info.setUInt64(numUnderrun);
            break;
        case STREAM_INFO_NUM_ANNOUNCED:
            info.setSizeT(buffers.size());
            break;
        case STREAM_INFO_NUM_QUEUED:
            info.setSizeT(inputPool.size());
            break;
        case STREAM_INFO_NUM_AWAIT_DELIVERY:
            info.setSizeT(outputQueue.size());
            break;
        case STREAM_INFO_NUM_STARTED:
            info.setUInt64(numCaptured);
            break;
        case STREAM_INFO_PAYLOAD_SIZE:
            info.setSizeT(getPayloadSize());
            break;
        case STREAM_INFO_IS_GRABBING:
            info.setBool(framesToAquire > 0);
            break;
        case STREAM_INFO_DEFINES_PAYLOADSIZE:
            info.setBool(true);
            break;
        case STREAM_INFO_TLTYPE:
            info.setString("Ethernet");
            break;
        case STREAM_INFO_NUM_CHUNKS_MAX:
            info.setSizeT(0);
        case STREAM_INFO_BUF_ANNOUNCE_MIN:
            info.setSizeT(1);
            break;
        case STREAM_INFO_BUF_ALIGNMENT:
            info.setSizeT(16);
            break;
        default:
            return GC_ERR_NOT_IMPLEMENTED;
    }

    return info.query();
}

GC_ERROR DataStream::getNumBufferParts(BUFFER_HANDLE hBuffer, uint32_t *piNumParts) {
    if(streamType == MULTIPART_STREAM) {
        // Multipart stream
        *piNumParts = 3;
    } else {
        *piNumParts = 1;
    }

    return GC_ERR_SUCCESS;
}

GC_ERROR DataStream::getBufferPartInfo(BUFFER_HANDLE hBuffer, uint32_t iPartIndex,
        BUFFER_PART_INFO_CMD iInfoCmd, INFO_DATATYPE *piType, void *pBuffer, size_t *piSize) {

    if(streamType != MULTIPART_STREAM) {
        // Part info is only available for multi part data streams
        return GC_ERR_NO_DATA;
    }

    Buffer* buffer = reinterpret_cast<Buffer*>(hBuffer);
    bool hasDisparityStream = (buffer->getMetaData().getPixelFormat(1) == ImagePair::FORMAT_12_BIT);

    StreamType streamType;
    switch(iPartIndex) {
        case 0:
            streamType = IMAGE_LEFT_STREAM;
            break;
        case 1:
            if(hasDisparityStream) {
                streamType = DISPARITY_STREAM;
            } else {
                streamType = IMAGE_RIGHT_STREAM;
            }
            break;
        case 2:
            streamType = POINTCLOUD_STREAM;
            break;
        default:
            return GC_ERR_INVALID_INDEX;
    }

    InfoQuery info(piType, pBuffer, piSize);
    switch(iInfoCmd) {
        case BUFFER_PART_INFO_BASE: {
            // Get the sizes of all stream parts before this one
            size_t offset = 0;
            if(streamType != IMAGE_LEFT_STREAM) {
                offset += getPayloadSizeForStreamType(IMAGE_LEFT_STREAM);

                if(streamType == POINTCLOUD_STREAM) {
                    offset += getPayloadSizeForStreamType(hasDisparityStream ? DISPARITY_STREAM : IMAGE_RIGHT_STREAM);
                }
            }

            info.setPtr(&buffer->getData()[offset]);
            break;
        }
        case BUFFER_PART_INFO_DATA_SIZE:
            info.setSizeT(getPayloadSizeForStreamType(streamType));
            break;
        case BUFFER_PART_INFO_DATA_TYPE:
            if(streamType != POINTCLOUD_STREAM) {
                info.setSizeT(PART_DATATYPE_2D_IMAGE);
            } else {
                info.setSizeT(PART_DATATYPE_3D_IMAGE);
            }
            break;
        case BUFFER_PART_INFO_DATA_FORMAT:
            info.setUInt64(getPixelFormatForStreamType(buffer->getMetaData(), streamType));
            break;
        case BUFFER_PART_INFO_DATA_FORMAT_NAMESPACE:
            info.setUInt64(PIXELFORMAT_NAMESPACE_PFNC_32BIT);
            break;
        case BUFFER_PART_INFO_WIDTH:
            info.setSizeT(buffer->getMetaData().getWidth());
            break;
        case BUFFER_PART_INFO_HEIGHT:
            info.setSizeT(buffer->getMetaData().getHeight());
            break;
        case BUFFER_PART_INFO_XOFFSET:
            info.setSizeT(0);
            break;
        case BUFFER_PART_INFO_YOFFSET:
            info.setSizeT(0);
            break;
        case BUFFER_PART_INFO_XPADDING:
            info.setSizeT(0);
            break;
        case BUFFER_PART_INFO_SOURCE_ID:
            info.setUInt64(0);
            break;
        case BUFFER_PART_INFO_DELIVERED_IMAGEHEIGHT:
            info.setSizeT(buffer->getMetaData().getHeight());
            break;
        default:
            return GC_ERR_NOT_IMPLEMENTED;
    }

    return info.query();
}

}
